@@ -6,18 +6,24 @@
 # the Apache License 2.0.  The full license can be found in the LICENSE file.
 #
 
+import re
+from random import random, randint, choices
+from string import ascii_letters
 from pytest import raises
 from datetime import date, datetime
 from time import mktime
 from perspective import Table, PerspectiveCppError
 from .test_view import compare_delta
 
+def randstr(length, input=ascii_letters):
+    return "".join(choices(input, k=length))
 
 class TestViewExpression(object):
     def test_table_validate_expressions_empty(self):
         table = Table({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
         validate = table.validate_expressions([])
         assert validate["expression_schema"] == {}
+        assert validate["expression_alias"] == {}
         assert validate["errors"] == {}
 
     def test_view_expression_schema_empty(self):
@@ -26,12 +32,99 @@ class TestViewExpression(object):
         assert view.to_columns() == {"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]}
         assert view.expression_schema() == {}
 
+    def test_view_validate_expressions_alias_map_errors(self):
+        table = Table({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
+        expressions = [
+            '// x\n"a"',
+            '// y\n"b" * 0.5',
+            "// c\n'abcdefg'",
+            "// d\ntrue and false",
+            '// e\nfloat("a") > 2 ? null : 1',
+            "// f\ntoday()",
+            "// g\nnow()",
+            "// h\nlength(123)",
+        ]
+
+        validated = table.validate_expressions(expressions)
+
+        aliases = ["x", "y", "c", "d", "e", "f", "g", "h"]
+
+        # Errored should also be in aliases
+        for idx, alias in enumerate(aliases):
+            assert validated["expression_alias"][alias] == expressions[idx]
+
+    def test_view_validate_expressions_alias_map(self):
+        table = Table({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
+        expressions = [
+            '// x\n"a"',
+            '// y\n"b" * 0.5',
+            "// c\n'abcdefg'",
+            "// d\ntrue and false",
+            '// e\nfloat("a") > 2 ? null : 1',
+            "// f\ntoday()",
+            "// g\nnow()",
+            "// h\nlength('abcd')",
+        ]
+
+        validated = table.validate_expressions(expressions)
+
+        aliases = ["x", "y", "c", "d", "e", "f", "g", "h"]
+
+        for idx, alias in enumerate(aliases):
+            assert validated["expression_alias"][alias] == expressions[idx]
+
+    def test_view_expression_schema_all_types(self):
+        table = Table({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
+        expressions = [
+            '"a"',
+            '"b" * 0.5',
+            "'abcdefg'",
+            "true and false",
+            'float("a") > 2 ? null : 1',
+            "today()",
+            "now()",
+            "length('abcd')",
+        ]
+
+        view = table.view(expressions=expressions)
+        assert view.expression_schema() == {
+            '"a"': int,
+            '"b" * 0.5': float,
+            "'abcdefg'": str,
+            "true and false": bool,
+            'float("a") > 2 ? null : 1': float,
+            "today()": date,
+            "now()": datetime,
+            "length('abcd')": float,
+        }
+
+        result = view.to_columns()
+        today = datetime(date.today().year, date.today().month, date.today().day)
+        del result["now()"]  # no need to match datetime.now()
+
+        assert result == {
+            "a": [1, 2, 3, 4],
+            "b": [5, 6, 7, 8],
+            '"a"': [1, 2, 3, 4],
+            '"b" * 0.5': [2.5, 3, 3.5, 4],
+            "'abcdefg'": ['abcdefg' for _ in range(4)],
+            "true and false": [False for _ in range(4)],
+            'float("a") > 2 ? null : 1': [1, 1, None, None],
+            "today()": [today for _ in range(4)],
+            "length('abcd')": [4 for _ in range(4)]
+        }
+
+        validated = table.validate_expressions(expressions)
+
+        for expr in expressions:
+            assert validated["expression_alias"][expr] == expr
+
     def test_table_validate_expressions_with_errors(self):
         table = Table({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
-        validate = table.validate_expressions(
-            ['"Sales" + "Profit"', "datetime()", "string()", "for () {}"]
-        )
+        expressions = ['"Sales" + "Profit"', "datetime()", "string()", "for () {}"]
+        validate = table.validate_expressions(expressions)
         assert validate["expression_schema"] == {}
+        assert validate["expression_alias"] == {expr: expr for expr in expressions}
         assert validate["errors"] == {
             '"Sales" + "Profit"': {
                 "column": 0,
@@ -65,6 +158,340 @@ class TestViewExpression(object):
         }
         assert view.expression_schema() == {"computed": float}
 
+    def test_view_expression_string_per_page(self):
+        table = Table({"a": [i for i in range(100)]})
+        big_strings = [randstr(6400) for _ in range(4)]
+        view = table.view(
+            expressions=[
+                "//computed{}\nvar x := '{}'; lower(x)".format(i, big_strings[i]) for i in range(4)
+            ]
+        )
+
+        result = view.to_columns()
+        schema = view.expression_schema()
+
+        for i in range(4):
+            name = "computed{}".format(i)
+            res = big_strings[i].lower()
+            assert schema[name] == str
+            assert result[name] == [res for _ in range(100)]
+
+    def test_view_expression_string_page_stress(self):
+        table = Table({"a": [i for i in range(100)]})
+        big_strings = [
+            "".join(["a" for _ in range(640)]),
+            "".join(["b" for _ in range(640)]),
+            "".join(["c" for _ in range(640)]),
+            "".join(["d" for _ in range(640)]),
+        ]
+
+        view = table.view(
+            expressions=[
+                "//computed\nvar a := '{}'; var b := '{}'; var c := '{}'; var d := '{}'; concat(a, b, c, d)".format(*big_strings)
+            ]
+        )
+
+        result = view.to_columns()
+        schema = view.expression_schema()
+        assert schema == {"computed": str}
+        assert result["computed"] == ["".join(big_strings) for _ in range(100)]
+
+    def test_view_expression_new_vocab_page(self):
+        table = Table({"a": [randstr(100) for _ in range(100)]})
+
+        def make_expression(idx):
+            expr = ["//computed{}".format(idx)]
+            num_vars = randint(1, 26)
+            concat_cols = []
+            concat_result = []
+
+            for i in range(num_vars):
+                name = ascii_letters[i]
+                string_literal = randstr(randint(100, 1000))
+
+                if random() > 0.5:
+                    result = string_literal.upper()
+                    string_literal = "upper('{}')".format(string_literal)
+                else:
+                    result = string_literal.lower()
+                    string_literal = "lower('{}')".format(string_literal)
+
+                concat_cols.append(name)
+                concat_result.append(result)
+
+                expr.append("var {} := {};".format(name, string_literal))
+
+            expr.append("concat(\"a\", {})".format(", ".join(concat_cols)))
+
+            return {
+                "expression_name": expr[0][2:],
+                "expression": "\n".join(expr),
+                "output": "".join(concat_result)
+            }
+
+        expressions = [make_expression(i) for i in range(10)]
+
+        view = table.view(
+            expressions=[expr["expression"] for expr in expressions]
+        )
+
+        result = view.to_columns()
+        schema = view.expression_schema()
+
+        for expr in expressions:
+            name = expr["expression_name"]
+            assert schema[name] == str
+
+            for i in range(100):
+                val = result["a"][i]
+                assert result[name][i] == val + expr["output"]
+
+    def test_view_expression_collide_local_var(self):
+        """Make sure that strings declared under the same var name in
+        different expressions do not collide."""
+        table = Table({"a": [1, 2, 3, 4]})
+        strings = [randstr(50) for _ in range(8)]
+
+        view = table.view(
+            expressions=[
+                "// computed \n var w := '{}'; var x := '{}'; var y := '{}'; var z := '{}'; concat(w, x, y, z)".format(*strings[:4]),
+                "// computed2 \n var w := '{}'; var x := '{}'; var y := '{}'; var z := '{}'; concat(w, x, y, z)".format(*strings[4:]),
+            ]
+        )
+
+        result = view.to_columns()
+        schema = view.expression_schema()
+
+        assert schema == {
+            "computed": str,
+            "computed2": str
+        }
+
+        assert result["computed"] == ["".join(strings[:4]) for _ in range(4)]
+        assert result["computed2"] == ["".join(strings[4:]) for _ in range(4)]
+
+    def test_view_random_expressions(self):
+        def make_expression():
+            """Create a random expression with a few local string vars that
+            are too long to be stored in-place."""
+            expression_name = randstr(10)
+            expression = ["// {}\n".format(expression_name)]
+            num_vars = randint(1, 26)
+            output_var_name = ""
+            output_str = ""
+
+            for i in range(num_vars):
+                name = ascii_letters[i]
+                string_literal = randstr(randint(15, 100))
+                expression.append("var {} := '{}';\n".format(name, string_literal))
+
+                if i == num_vars - 1:
+                    output_var_name = name
+                    output_str = string_literal
+
+            expression.append(output_var_name)
+            return {
+                "expression_name": expression_name,
+                "expression": "".join(expression),
+                "output": output_str
+            }
+
+        table = Table({"a": [1, 2, 3, 4]})
+
+        for _ in range(5):
+            exprs = [make_expression() for _ in range(5)]
+            output_map = {expr["expression_name"]: expr["output"] for expr in exprs}
+            view = table.view(expressions=[expr["expression"] for expr in exprs])
+            expression_schema = view.expression_schema()
+            result = view.to_columns()
+
+            for expr in output_map.keys():
+                assert expression_schema[expr] == str
+                assert result[expr] == [output_map[expr] for _ in range(4)]
+
+    def test_view_expression_string_literal_compare(self):
+        table = Table({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
+        validated = table.validate_expressions(['// computed \n \'a\' == \'a\''])
+
+        assert validated["expression_schema"] == {
+            "computed": "boolean"
+        }
+
+        view = table.view(expressions=['// computed \n \'a\' == \'a\''])
+
+        assert view.to_columns() == {
+            "a": [1, 2, 3, 4],
+            "b": [5, 6, 7, 8],
+            "computed": [True, True, True, True],
+        }
+
+        assert view.expression_schema() == {"computed": bool}
+
+    def test_view_expression_string_literal_compare_null(self):
+        table = Table({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
+        validated = table.validate_expressions(['// computed \n \'a\' == null'])
+
+        assert validated["expression_schema"] == {
+            "computed": "float"
+        }
+
+        view = table.view(expressions=['// computed \n \'a\' == null'])
+
+        assert view.to_columns() == {
+            "a": [1, 2, 3, 4],
+            "b": [5, 6, 7, 8],
+            "computed": [0, 0, 0, 0],
+        }
+
+        assert view.expression_schema() == {"computed": float}
+
+    def test_view_expression_string_literal_compare_column(self):
+        table = Table({"a": ["a", "a", "b", "c"]})
+        validated = table.validate_expressions(['// computed \n "a" == \'a\''])
+
+        assert validated["expression_schema"] == {
+            "computed": "boolean"
+        }
+
+        view = table.view(expressions=['// computed \n "a" == \'a\''])
+
+        assert view.to_columns() == {
+            "a": ["a", "a", "b", "c"],
+            "computed": [True, True, False, False],
+        }
+
+        assert view.expression_schema() == {"computed": bool}
+
+    def test_view_expression_string_literal_compare_column_null(self):
+        table = Table({"a": ["a", None, "b", "c", None]})
+        validated = table.validate_expressions(['// computed \n "a" == \'a\''])
+
+        assert validated["expression_schema"] == {
+            "computed": "boolean"
+        }
+
+        view = table.view(expressions=['// computed \n "a" == \'a\''])
+
+        assert view.to_columns() == {
+            "a": ["a", None, "b", "c", None],
+            "computed": [True, False, False, False, False],
+        }
+
+        assert view.expression_schema() == {"computed": bool}
+
+    def test_view_expression_string_literal_compare_column_null_long(self):
+        table = Table({"a": ["abcdefghijklmnopqrstuvwxyz", None, "abcdefghijklmnopqrstuvwxyz", "aabcdefghijklmnopqrstuvwxyz", None]})
+        validated = table.validate_expressions(['// computed \n "a" == \'abcdefghijklmnopqrstuvwxyz\''])
+
+        assert validated["expression_schema"] == {
+            "computed": "boolean"
+        }
+
+        view = table.view(expressions=['// computed \n "a" == \'abcdefghijklmnopqrstuvwxyz\''])
+        result = view.to_columns()
+        assert result["computed"] == [True, False, True, False, False]
+
+        assert view.expression_schema() == {"computed": bool}
+
+    def test_view_expression_string_literal_compare_column_null_long_var(self):
+        table = Table({"a": ["abcdefghijklmnopqrstuvwxyz", None, "abcdefghijklmnopqrstuvwxyz", "aabcdefghijklmnopqrstuvwxyz", None]})
+        validated = table.validate_expressions(['// computed \n var xyz := \'abcdefghijklmnopqrstuvwxyz\'; "a" == xyz'])
+
+        assert validated["expression_schema"] == {
+            "computed": "boolean"
+        }
+
+        view = table.view(expressions=['// computed \n var xyz := \'abcdefghijklmnopqrstuvwxyz\'; "a" == xyz'])
+        result = view.to_columns()
+        assert result["computed"] == [True, False, True, False, False]
+        assert view.expression_schema() == {"computed": bool}
+
+    def test_view_expression_string_literal_compare_if(self):
+        table = Table({"a": ["a", "a", "b", "c"]})
+        validated = table.validate_expressions(['// computed \n if("a" == \'a\', 1, 2)'])
+
+        assert validated["expression_schema"] == {
+            "computed": "float"
+        }
+
+        view = table.view(expressions=['// computed \n if("a" == \'a\', 1, 2)'])
+
+        assert view.to_columns() == {
+            "a": ["a", "a", "b", "c"],
+            "computed": [1, 1, 2, 2],
+        }
+
+        assert view.expression_schema() == {"computed": float}
+
+    def test_view_expression_string_literal_var(self):
+        table = Table({"a": [1, 2, 3]})
+
+        for _ in range(10):
+            view = table.view(expressions=["var x := 'Eabcdefghijklmn'; var y := '0123456789'; concat(x, y)"])
+            assert view.to_columns() == {
+                "a": [1, 2, 3],
+                "var x := 'Eabcdefghijklmn'; var y := '0123456789'; concat(x, y)": ["Eabcdefghijklmn0123456789", "Eabcdefghijklmn0123456789", "Eabcdefghijklmn0123456789"]
+            }
+
+    def test_view_streaming_expression(self):
+        def data():
+            return [{"a": random()} for _ in range(50)]
+
+        table = Table(data())
+        view = table.view(expressions=["123"])
+
+        for _ in range(5):
+            table.update(data())
+        
+        assert table.size() == 300
+        result = view.to_dict()
+        assert result["123"] == [123 for _ in range(300)]
+
+    def test_view_streaming_expression_limit(self):
+        def data():
+            return [{"a": random()} for _ in range(55)]
+
+        table = Table(data(), limit=50)
+        view = table.view(expressions=["123"])
+
+        for _ in range(5):
+            table.update(data())
+        
+        assert table.size() == 50
+        result = view.to_dict()
+        assert result["123"] == [123 for _ in range(50)]
+
+    def test_view_streaming_expression_one(self):
+        def data():
+            return [{"a": random()} for _ in range(50)]
+
+        table = Table(data())
+        view = table.view(row_pivots=["c0"], expressions=['//c0\n"a" * 2'])
+
+        for _ in range(5):
+            table.update(data())
+        
+        assert table.size() == 300
+        assert view.expression_schema() == {
+            "c0": float
+        }
+
+    def test_view_streaming_expression_two(self):
+        def data():
+            return [{"a": random()} for _ in range(50)]
+
+        table = Table(data())
+        view = table.view(row_pivots=["c0"], column_pivots=["c1"], expressions=['//c0\n"a" * 2', "//c1\n'new string'"])
+
+        for i in range(5):
+            table.update(data())
+        
+        assert table.size() == 300
+        assert view.expression_schema() == {
+            "c0": float,
+            "c1": int  # pivoted
+        }
+
     def test_view_expression_create_no_alias(self):
         table = Table({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
         view = table.view(expressions=['"a" + "b"'])
@@ -1243,3 +1670,111 @@ class TestViewExpression(object):
         assert result["computed5"] == ["abcdefghijklmnop", "def"]
         assert result["computed6"] == ["false", "true"]
         assert result["computed7"] == ["1234.57"] * 2
+
+    def test_view_expession_multicomment(self):
+        table = Table({"a": [1, 2, 3, 4]})
+        view = table.view(expressions=["//abc\nvar x := 1 + 2; // def\nx + 100 // cdefghijk"])
+        assert view.expression_schema() == {"abc": float}
+        assert view.to_columns() == {
+            "abc": [103, 103, 103, 103],
+            "a": [1, 2, 3, 4]
+        }
+    
+    def test_view_regex_email(self):
+        endings = ["com", "net", "co.uk", "ie", "me", "io", "co"]
+        data = ["{}@{}.{}".format(randstr(30, ascii_letters + "0123456789" + "._-"), randstr(10), choices(endings, k=1)[0]) for _ in range(100)]
+        table = Table({"a": data})
+        expressions = [
+            "// address\nsearch(\"a\", '^([a-zA-Z0-9._-]+)@')",
+            "// domain\nsearch(\"a\", '@([a-zA-Z.]+)$')",
+            "//is_email?\nfullmatch(\"a\", '^([a-zA-Z0-9._-]+)@([a-zA-Z.]+)$')",
+            "//has_at?\nmatch(\"a\", '@')"
+        ]
+
+        view = table.view(expressions=expressions)
+        schema = view.expression_schema()
+        assert schema == {
+            "address": str,
+            "domain": str,
+            "is_email?": bool,
+            "has_at?": bool
+        }
+
+        results = view.to_columns()
+
+        for i in range(100):
+            source = results["a"][i]
+            expected_address = re.match(r"^([a-zA-Z0-9._-]+)@", source).group(1)
+            expected_domain = re.search(r"@([a-zA-Z.]+)$", source).group(1)
+            assert results["address"][i] == expected_address
+            assert results["domain"][i] == expected_domain
+            assert results["is_email?"][i] == True
+            assert results["has_at?"][i] == True
+
+    def test_view_expression_number(self):
+        def digits():
+            return randstr(4, "0123456789")
+        
+        data = []
+
+        for _ in range(1000):
+            separator = "-" if random() > 0.5 else " "
+            data.append("{}{}{}{}{}{}{}".format(digits(), separator, digits(), separator, digits(), separator, digits()))
+        
+        table = Table({"a": data})
+        view = table.view(expressions=["""// parsed\n
+        var parts[4];
+        parts[0] := search("a", '^([0-9]{4})[ -][0-9]{4}[ -][0-9]{4}[ -][0-9]{4}');
+        parts[1] := search("a", '^[0-9]{4}[ -]([0-9]{4})[ -][0-9]{4}[ -][0-9]{4}');
+        parts[2] := search("a", '^[0-9]{4}[ -][0-9]{4}[ -]([0-9]{4})[ -][0-9]{4}');
+        parts[3] := search("a", '^[0-9]{4}[ -][0-9]{4}[ -][0-9]{4}[ -]([0-9]{4})');
+        concat(parts[0], parts[1], parts[2], parts[3])
+        """, "//is_number?\nfullmatch(\"a\", '^[0-9]{4}[ -][0-9]{4}[ -][0-9]{4}[ -][0-9]{4}')"])
+        schema = view.expression_schema()
+        assert schema == {"parsed": str, "is_number?": bool}
+        results = view.to_columns()
+        
+        for i in range(1000):
+            source = results["a"][i]
+            expected = re.sub(r"[ -]", "", source)
+            assert results["parsed"][i] == expected
+            assert results["is_number?"][i] == True
+
+
+    def test_view_expression_newlines(self):
+        table = Table({"a": [
+            "abc\ndef",
+            "\n\n\n\nabc\ndef",
+            "abc\n\n\n\n\n\nabc\ndef\n\n\n\n",
+            None,
+            "def",
+        ],
+        "b": [
+            "hello\tworld",
+            "\n\n\n\n\nhello\n\n\n\n\n\tworld",
+            "\tworld",
+            "world",
+            None,
+        ]})
+
+        view = table.view(
+            expressions=[
+                "//c1\nsearch(\"a\", '(\ndef)')",
+                "//c2\nsearch(\"b\", '(\tworld)')",
+                "//c3\nmatch(\"a\", '\\n')",
+                "//c4\nmatch(\"b\", '\\n')"
+            ]
+        )
+
+        assert view.expression_schema() == {
+            "c1": str,
+            "c2": str,
+            "c3": bool,
+            "c4": bool
+        }
+
+        results = view.to_columns()
+        assert results["c1"] == ["\ndef", "\ndef", "\ndef", None, None]
+        assert results["c2"] == ["\tworld", "\tworld", "\tworld", None, None]
+        assert results["c3"] == [True, True, True, None, False]
+        assert results["c4"] == [False, True, False, False, None]
